@@ -1,5 +1,5 @@
 ## This file is part of simpleserver
-## Copyright (C) 2000-2015 Index Data.
+## Copyright (C) 2000-2017 Index Data.
 ## All rights reserved.
 ## Redistribution and use in source and binary forms, with or without
 ## modification, are permitted provided that the following conditions are met:
@@ -36,7 +36,7 @@ require AutoLoader;
 
 @ISA = qw(Exporter AutoLoader DynaLoader);
 @EXPORT = qw( );
-$VERSION = '1.20';
+$VERSION = '1.27';
 
 bootstrap Net::Z3950::SimpleServer $VERSION;
 
@@ -99,6 +99,9 @@ sub launch_server {
 	if (defined($self->{START})) {
 		set_start_handler($self->{START});
 	}
+	if (defined($self->{ESREQUEST})) {
+		set_esrequest_handler($self->{ESREQUEST});
+	}
 	start_server(@args);
 }
 
@@ -178,9 +181,11 @@ package Net::Z3950::SimpleServer;
 __END__
 # Below is the stub of documentation for your module. You better edit it!
 
+=encoding utf8
+
 =head1 NAME
 
-Net::Z3950::SimpleServer - Simple Perl API for building Z39.50 servers. 
+Net::Z3950::SimpleServer - Simple Perl API for building Z39.50 servers.
 
 =head1 SYNOPSIS
 
@@ -216,10 +221,10 @@ Net::Z3950::SimpleServer - Simple Perl API for building Z39.50 servers.
 
   ## Register custom event handlers:
   my $z = new Net::Z3950::SimpleServer(GHANDLE = $someObject,
-				       INIT   =>  \&my_init_handler,
-				       CLOSE  =>  \&my_close_handler,
-				       SEARCH =>  \&my_search_handler,
-				       FETCH  =>  \&my_fetch_handler);
+				       INIT   =>  "main::my_init_handler",
+				       CLOSE  =>  "main::my_close_handler",
+				       SEARCH =>  "main::my_search_handler",
+				       FETCH  =>  "main::my_fetch_handler");
 
   ## Launch server:
   $z->launch_server("ztest.pl", @ARGV);
@@ -255,7 +260,7 @@ of events:
   - Search request
   - Present request
   - Fetching of records
-  - Scan request (browsing) 
+  - Scan request (browsing)
   - Closing down connection
 
 Note that only the Search and Fetch handler functions are required.
@@ -273,36 +278,22 @@ The Perl programmer specifies the event handlers for the server by
 means of the SimpleServer object constructor
 
   my $z = new Net::Z3950::SimpleServer(
-                        START   =>      \&my_start_handler,
-			INIT	=>	\&my_init_handler,
-			CLOSE	=>	\&my_close_handler,
-			SEARCH	=>	\&my_search_handler,
-			PRESENT	=>	\&my_present_handler,
-			SCAN	=>	\&my_scan_handler,
-			FETCH	=>	\&my_fetch_handler,
-  			EXPLAIN =>	\&my_explain_handler,
-  			DELETE  =>	\&my_delete_handler,
-  			SORT    =>	\&my_sort_handler);
+                        START   =>      "main::my_start_handler",
+			INIT	=>	"main::my_init_handler",
+			CLOSE	=>	"main::my_close_handler",
+			SEARCH	=>	"main::my_search_handler",
+			PRESENT	=>	"main::my_present_handler",
+			SCAN	=>	"main::my_scan_handler",
+			FETCH	=>	"main::my_fetch_handler",
+			EXPLAIN =>	"main::my_explain_handler",
+			DELETE  =>	"main::my_delete_handler",
+			ESREQUEST =>    "main::my_esrequest_handler",
+			SORT    =>	"main::my_sort_handler");
 
 In addition, the arguments to the constructor may include GHANDLE, a
 global handle which is made available to each invocation of every
 callback function.  This is typically a reference to either a hash or
-an object.
-
-If you want your SimpleServer to start a thread (threaded mode) to
-handle each incoming Z39.50 request instead of forking a process
-(forking mode), you need to register the handlers by symbol rather
-than by code reference. Thus, in threaded mode, you will need to
-register your handlers this way:
-
-  my $z = new Net::Z3950::SimpleServer(
-  			INIT	=>	"my_package::my_init_handler",
-			CLOSE	=>	"my_package::my_close_handler",
-			....
-			....          );
-
-where my_package is the Perl package in which your handler is
-located.
+an object. (replace main with your package, if not main).
 
 After the custom event handlers are declared, the server is launched
 by means of the method
@@ -312,7 +303,7 @@ by means of the method
 Notice, the first argument should be the name of your server
 script (for logging purposes), while the rest of the arguments
 are documented in the YAZ toolkit manual: The section on
-application invocation: <http://indexdata.com/yaz/doc/server.invocation.tkl>
+application invocation: <http://www.indexdata.com/yaz/doc/server.invocation.html>
 
 In particular, you need to use the -T switch to start your SimpleServer
 in threaded mode.
@@ -358,12 +349,14 @@ The argument hash passed to the init handler has the form
 	     			    ## this member contains user name
 	     PASS      =>  "yyy"    ## Under same conditions, this member
 	     			    ## contains the password in clear text
-	     GHANDLE   =>  $obj     ## Global handler specified at creation
+	     GROUP     =>  "zzz"    ## Under same conditions, this member
+				    ## contains the group in clear text
+	     GHANDLE   =>  $obj     ## Global handle specified at creation
 	     HANDLE    =>  undef    ## Handler of Perl data structure
 	  };
 
 The HANDLE member can be used to store any scalar value which will then
-be provided as input to all subsequent calls (ie. for searching, record
+be provided as input to all subsequent calls (i.e. for searching, record
 retrieval, etc.). A common use of the handle is to store a reference to
 a hash which may then be used to store session-specific parameters.
 If you have any session-specific information (such as a list of
@@ -382,27 +375,35 @@ associated additional information, ERR_STR returned.
 
 =head2 Search handler
 
-Similarly, the search handler is called with a reference to an anony-
-mous hash. The structure is the following:
+Similarly, the search handler is called with a reference to an
+anonymous hash. The structure is the following:
 
   $args = {
-	  			    ## Request parameters:
+                                     ## Request parameters:
 
-	     GHANDLE   =>  $obj     ## Global handler specified at creation
-	     HANDLE    =>  ref,     ## Your session reference.
-	     SETNAME   =>  "id",    ## ID of the result set
-	     REPL_SET  =>  0,       ## Replace set if already existing?
-	     DATABASES =>  ["xxx"], ## Reference to a list of data-
-				    ## bases to search
-	     QUERY     =>  "query", ## The query expression
-	     RPN       =>  $obj,    ## Reference to a Net::Z3950::APDU::Query
+    GHANDLE             =>  $obj     # Global handle specified at creation
+    HANDLE              =>  ref,     # Your session reference.
+    SETNAME             =>  "id",    # ID of the result set
+    REPL_SET            =>  0,       # Replace set if already existing?
+    DATABASES           =>  ["xxx"], # Reference to a list of databases to search
+    QUERY               =>  "query", # The query expression as a PQF string
+    RPN                 =>  $obj,    # Reference to a Net::Z3950::APDU::Query
+    CQL                 =>  $x,      # A CQL query, if this is provided instead of Type-1
+    SRW_SORTKEYS        =>  $x,      # XXX to be described
+    PID                 =>  $x,      # XXX to be described
+    PRESENT_NUMBER      =>  $x,      # XXX to be described
+    EXTRA_ARGS          =>  $x,      # XXX to be described
+    INPUTFACETS         =>  $x,      # Specification of facets required: see below.
 
-				    ## Response parameters:
+                                     ## Response parameters:
 
-	     ERR_CODE  =>  0,       ## Error code (0=Successful search)
-	     ERR_STR   =>  "",      ## Error string
-	     HITS      =>  0        ## Number of matches
-	  };
+    ERR_CODE            =>  0,       # Error code (0=Successful search)
+    ERR_STR             =>  "",      # Error string
+    HITS                =>  0,       # Number of matches
+    ESTIMATED_HIT_COUNT =>  $x,      # XXX to be described
+    EXTRA_RESPONSE_DATA =>  $x,      # XXX to be described
+    OUTPUTFACETS        =>  $x       # Facets returned: see below.
+  };
 
 Note that a search which finds 0 hits is considered successful in
 Z39.50 terms - you should only set the ERR_CODE to a non-zero value
@@ -410,13 +411,17 @@ if there was a problem processing the request. The Z39.50 standard
 provides a comprehensive list of standard diagnostic codes, and you
 should use these whenever possible.
 
-The QUERY is a tree-structure of terms combined by operators, the
-terms being qualified by lists of attributes. The query is presented
-to the search function in the Prefix Query Format (PQF) which is
-used in many applications based on the YAZ toolkit. The full grammar
-is described in the YAZ manual.
+=head3 Query structures
 
-The following are all examples of valid queries in the PQF. 
+In Z39.50, the most comment kind of query is the so-called Type-1
+_query, a tree-structure of terms combined by operators, the terms
+being qualified by lists of attributes.
+
+The QUERY parameter presented this tree to the search function in the
+Prefix Query Format (PQF) which is used in many applications based on
+the YAZ toolkit. The full grammar is described in the YAZ manual.
+
+The following are all examples of valid queries in the PQF.
 
 	dylan
 
@@ -435,12 +440,12 @@ The following are all examples of valid queries in the PQF.
 You will need to write a recursive function or something similar to
 parse incoming query expressions, and this is usually where a lot of
 the work in writing a database-backend happens. Fortunately, you don't
-need to support anymore functionality than you want to. For instance,
-it is perfectly legal to not accept boolean operators, but you SHOULD
+need to support any more functionality than you want to. For instance,
+it is perfectly legal to not accept boolean operators, but you should
 try to return good error codes if you run into something you can't or
 won't support.
 
-A more convenient alternative to the QUERY member may be the RPN
+A more convenient alternative to the QUERY member is the RPN
 member, which is a reference to a Net::Z3950::APDU::Query object
 representing the RPN query tree.  The structure of that object is
 supposed to be self-documenting, but here's a brief summary of what
@@ -483,11 +488,11 @@ Z<>
 =item C<Net::Z3950::RPN::AndNot>
 
 These three classes are all arrays of two elements, each of which is a
-node of one of the above types.
+node.
 
 =item C<Net::Z3950::RPN::Term>
 
-See below for details.
+A query term. See below for details.
 
 =item C<Net::Z3950::RPN::RSID>
 
@@ -574,6 +579,168 @@ By adding your own methods to these classes (building what I call
 of the incoming RPN.  Take a look at C<samples/render-search.pl> for a
 sample implementation of such an augmented classes technique.
 
+Finally, when SimpleServer is invoked using SRU/SRW (and indeed using
+Z39.50 if the unusual type-104 query is used), the query that is
+_passed is expressed in CQL, the Contextual Query Language. In this
+case, the query string is made available in the CQL argument.
+
+=head3 Facets
+
+Servers may support the provision of facets -- counted lists of field
+values which may subsequently be be used as query terms to narrow the
+search.
+
+In SRU, facets may be requested by the C<facetLimit> parameter,
+L<as documented in the OASIS standard that formalises the SRU specification|http://docs.oasis-open.org/search-ws/searchRetrieve/v1.0/os/part3-sru2.0/searchRetrieve-v1.0-os-part3-sru2.0.html#_Toc324162453>.
+Its value is a string consisting of a comma-separated list of facet
+specifications. Each facet specification consists of of a count, a
+colon and a fieldname. For example, C<facetLimit=10:title,5:author>
+asks for ten title facets and five author facets.
+
+=head4 Request format
+
+The facet request is passed to the search-handler function in the
+INPUTFACETS parameter. Its value is rather complex, due to backwards
+compatibility with Z39.50:
+
+=over 4
+
+=item *
+
+The top-level value is a C<Net::Z3950::FacetList> array.
+
+=item *
+
+This is an array of C<Net::Z3950::FacetField> objects.
+
+=item *
+
+Each of these is an object with two members, C<attributes> and
+C<terms>.
+
+=item *
+
+C<attributes> has type C<Net::Z3950::RPN::Attributes> and is a list of
+objects of type C<Net::Z3950::RPN::Attribute>.
+
+=item *
+
+Each attribute has two elements, C<attributeType> and
+C<attributeValue>. Each value is interpreted according to its
+type. The meanings of the types are as follows:
+
+=over 4
+
+=item 1
+
+The name of the field to provide values of the facets.
+
+=item 2
+
+The order in which to sort the values. (But it's not clear how this is
+to be interpreted: it may be implementation dependent.)
+
+=item 3
+
+The number of facets to include for the specified field.
+
+=item 4
+
+The first facet to include in the response: for example, if this is
+11, then the first ten facts should be skipped.
+
+=back
+
+=back
+
+So for example, the SRU facet specification
+C<facetLimit=10:title,5:author> would be translated as a
+C<Net::Z3950::FacetList> list of two C<Net::Z3950::FacetField>s. The
+C<attributes> of the first would be [1="title", 3=10], and those of
+the second would be [1="author", 3=5].
+
+It is not clear what the purpose of C<terms> is, but for the record,
+this is how it is represented:
+
+=over 4
+
+=item *
+
+C<terms> is a C<Net::Z3950::FacetTerms> array.
+
+=item *
+
+This is an array of C<Net::Z3950::FacetTerm> objects.
+
+=item *
+
+Each of these is an object with two members, C<term> and C<count>. The
+first of these is an integer, the second a string.
+
+=back
+
+=head4 Response format
+
+Having generated facets corresponding to the request, the search
+handler should return them in the C<OUTPUTFACETS> argument. The
+structure of this response is similar to that of the request:
+
+=over 4
+
+=item *
+
+The top-level value is a C<Net::Z3950::FacetList> array.
+
+=item *
+
+This is an array of C<Net::Z3950::FacetField> objects.
+
+=item *
+
+Each of these is an object with two members, C<attributes> and
+C<terms>.
+
+=item *
+
+C<attributes> has type C<Net::Z3950::RPN::Attributes> and is a list of
+objects of type C<Net::Z3950::RPN::Attribute>.
+
+=item *
+
+Each attribute has two elements, C<attributeType> and
+C<attributeValue>. Each value is interpreted according to its
+type. The meanings of the types are as follows:
+
+=over 4
+
+=item 1
+
+The name of the field for which terms are provided.
+
+=back
+
+(That is the only type used.)
+
+=item *
+
+C<terms> is a C<Net::Z3950::FacetTerms> array.
+
+=item *
+
+This is an array of C<Net::Z3950::FacetTerm> objects.
+
+=item *
+
+Each of these is an object with two members, C<term> and C<count>. The
+first of these is a string containing one of the facet terms, and the
+second is an integer indicating how many times it occurs in the
+records that were found by the search.
+
+=back
+
+The example SimpleServer applicaation server C<ztest.pl> includes code
+that shows how to examine the INPUTFACETS data structure and create
+the OUTPUTFACETS structure.
 
 =head2 Present handler
 
@@ -600,11 +767,12 @@ The information exchanged between client and present handle is:
   $args = {
 				    ## Client/server request:
 
-	     GHANDLE   =>  $obj     ## Global handler specified at creation
+	     GHANDLE   =>  $obj     ## Global handle specified at creation
 	     HANDLE    =>  ref,     ## Reference to datastructure
 	     SETNAME   =>  "id",    ## Result set ID
 	     START     =>  xxx,     ## Start position
 	     COMP      =>  "",	    ## Desired record composition
+	     SCHEMA_OID => "",      ## Z39.50 schema (OID), if any
 	     NUMBER    =>  yyy,	    ## Number of requested records
 
 
@@ -622,18 +790,19 @@ The fetch handler is asked to retrieve a SINGLE record from a given
 result set (the front-end server will automatically call the fetch
 handler as many times as required).
 
-The parameters exchanged between the server and the fetch handler are
+The parameters exchanged between the server and the fetch handler are:
 
   $args = {
 				    ## Client/server request:
 
-	     GHANDLE   =>  $obj     ## Global handler specified at creation
+	     GHANDLE   =>  $obj     ## Global handle specified at creation
 	     HANDLE    =>  ref	    ## Reference to data structure
 	     SETNAME   =>  "id"     ## ID of the requested result set
 	     OFFSET    =>  nnn      ## Record offset number
 	     REQ_FORM  =>  "n.m.k.l"## Client requested format OID
 	     COMP      =>  "xyz"    ## Formatting instructions
-	     SCHEMA    =>  "abc"    ## Requested schema, if any
+	     SCHEMA_OID => "",      ## Z39.50 schema (OID), if any
+	     SCHEMA    =>  "abc"    ## Requested schema (string), if any
 
 				    ## Handler response:
 
@@ -647,20 +816,20 @@ The parameters exchanged between the server and the fetch handler are
 	     SCHEMA    =>  "abc"    ## Provided schema, if any
 	  };
 
-The REP_FORM value has by default the REQ_FORM value but can be set to
+The REP_FORM value has by default the REQ_FORM value, but can be set to
 something different if the handler desires. The BASENAME value should
 contain the name of the database from where the returned record originates.
 The ERR_CODE and ERR_STR works the same way they do in the search
 handler. If there is an error condition, the SUR_FLAG is used to
 indicate whether the error condition pertains to the record currently
 being retrieved, or whether it pertains to the operation as a whole
-(eg. the client has specified a result set which does not exist.)
+(e.g. the client has specified a result set which does not exist.)
 
 If you need to return USMARC records, you might want to have a look at
 the MARC module on CPAN, if you don't already have a way of generating
 these.
 
-NOTE: The record offset is 1-indexed - 1 is the offset of the first
+NOTE: The record offset is 1-indexed, so 1 is the offset of the first
 record in the set.
 
 =head2 Scan handler
@@ -671,17 +840,18 @@ should return an ordered list of specified length consisting of terms
 actually occurring in the data base. Each of these terms should be close
 to or equal to the term originally specified. The quality of scan compared
 to simple search is a guarantee of hits. It is simply like browsing through
-an index of a book, you always find something! The parameters exchanged are
+an index of a book, you always find something! The parameters exchanged are:
 
   $args = {
 						## Client request
 
-		GHANDLE		=> $obj,	## Global handler specified at creation
+		GHANDLE		=> $obj,	## Global handle specified at creation
 		HANDLE		=> $ref,	## Reference to data structure
 		DATABASES	=> ["xxx"],	## Reference to a list of data-
 						## bases to search
 		TERM		=> 'start',	## The start term
 		RPN		=>  $obj,       ## Reference to a Net::Z3950::RPN::Term
+		attributeSet	=> OID,		## OID String (optional)
 
 		NUMBER		=> xx,		## Number of requested terms
 		POS		=> yy,		## Position of starting point
@@ -739,7 +909,7 @@ The argument hash received by the close handler has two elements only:
   $args = {
 				    ## Server provides:
 
-	     GHANDLE   =>  $obj     ## Global handler specified at creation
+	     GHANDLE   =>  $obj     ## Global handle specified at creation
 	     HANDLE    =>  ref      ## Reference to data structure
 	  };
 
@@ -747,13 +917,30 @@ What ever data structure the HANDLE value points at goes out of scope
 after this call. If you need to close down a connection to your server
 or something similar, this is the place to do it.
 
+=head2 Explain handler
+
+The argument hash received by the explain handler has the following elements:
+
+  $args = {
+                            ## Request parameters:
+     GHANDLE   =>  $obj,    # Global handle specified at creation
+     HANDLE    =>  ref,     # Reference to data structure
+     DATABASE  =>  $dbname, # Name of database to be explained
+
+                            ## Response parameters:
+     EXPLAIN   =>  $zeerex  # ZeeRex record for specified database
+  };
+
+The handler should return a string containing the ZeeRex XML that
+describes that nominated database.
+
 =head2 Delete handler
 
 The argument hash received by the delete handler has the following elements:
 
   $args = {
 				    ## Client request:
-	     GHANDLE   =>  $obj,    ## Global handler specified at creation
+	     GHANDLE   =>  $obj,    ## Global handle specified at creation
 	     HANDLE    =>  ref,     ## Reference to data structure
 	     SETNAME   =>  "id",    ## Result set ID
 
@@ -768,7 +955,7 @@ be deleted.  In either case, the callback function should report on
 success or failure by setting the STATUS element either to zero, on
 success, or to an integer from 1 to 10, to indicate one of the ten
 possible failure codes described in section 3.2.4.1.4 of the Z39.50
-standard -- see 
+standard -- see
 http://www.loc.gov/z3950/agency/markup/05.html#Delete-list-statuses1
 
 =head2 Sort handler
@@ -777,7 +964,7 @@ The argument hash received by the sort handler has the following elements:
 
 	$args = {
 					## Client request:
-		GHANDLE => $obj,	## Global handler specified at creation
+		GHANDLE => $obj,	## Global handle specified at creation
 		HANDLE => ref,		## Reference to data structure
 		INPUT => [ a, b ... ],	## Names of result-sets to sort
 		OUTPUT => "name",	## Name of result-set to sort into
@@ -856,7 +1043,7 @@ and SRW protocols as well as Z39.50.  These ``web-friendly'' protocols
 enable similar functionality to that of Z39.50, but by means of rich
 URLs in the case of SRU, and a SOAP-based web-service in the case of
 SRW.  These protocols are described at
-http://www.loc.gov/sru
+http://www.loc.gov/standards/sru/
 
 In order to serve these protocols from a SimpleServer-based
 application, it is necessary to launch the application with a YAZ
@@ -875,7 +1062,7 @@ translate the CQL queries of SRU and SRW into corresponding Z39.50
 Type-1 queries.  For more information about YAZ GFS configuration,
 including how to specify an Explain record, see the I<Virtual Hosts>
 section of the YAZ manual at
-http://indexdata.com/yaz/doc/server.vhosts.tkl
+http://www.indexdata.com/yaz/doc/server.vhosts.html
 
 The mapping of CQL queries into Z39.50 Type-1 queries is specified by
 a file that indicates which BIB-1 attributes should be generated for
@@ -893,9 +1080,9 @@ Core indexes C<title>, C<subject> and C<creator>, and the BIB-1
 relations (type=2) corresponding to the CQL relations C<E<lt>> and
 C<E<lt>=>.  For more information about the format of this file, see
 the I<CQL> section of the YAZ manual at
-http://indexdata.com/yaz/doc/tools.tkl#tools.cql
+http://www.indexdata.com/yaz/doc/tools.html#cql
 
-The YAZ distribution include a sample CQL-to-PQF mapping configuration
+The YAZ distribution includes a sample CQL-to-PQF mapping configuration
 file called C<pqf.properties>; this is sufficient for many
 applications, and a good base to work from for most others.
 
@@ -911,12 +1098,12 @@ http://search.cpan.org/~esummers/CQL-Parser/
 =head1 AUTHORS
 
 Anders Sønderberg (sondberg@indexdata.dk),
-Sebastian Hammer (quinn@indexdata.dk),
+Sebastian Hammer (quinn@indexdata.com),
 Mike Taylor (indexdata.com).
 
 =head1 COPYRIGHT AND LICENCE
 
-Copyright (C) 2000-2015 by Index Data.
+Copyright (C) 2000-2017 by Index Data.
 
 This library is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself, either Perl version 5.8.4 or,
